@@ -11,14 +11,14 @@
 
 """
 
+import sys
 import re
 
 
-try:
+if sys.version_info[0] < 3:
     _unicode = unicode
     _unichr = unichr
-except NameError:
-    # Python 3
+else:
     _unicode = str
     _unichr = chr
 
@@ -43,7 +43,7 @@ class Selector(object):
     Represents a selector with an optional pseudo element.
     """
     def __init__(self, tree, pseudo_element=None):
-        self._tree = tree
+        self.parsed_tree = tree
         #: If the selector has a pseudo-element: a string like ``'after'``.
         #: Otherwise, ``None``.
         #: Any identifier preceded by ``::`` is accepted as a pseudo-element.
@@ -57,7 +57,7 @@ class Selector(object):
         else:
             pseudo_element = ''
         return '%s[%r%s]' % (
-            self.__class__.__name__, self._tree, pseudo_element)
+            self.__class__.__name__, self.parsed_tree, pseudo_element)
 
     def specificity(self):
         """Return the specificity_ of this selector as a tuple of 3 integers.
@@ -65,7 +65,7 @@ class Selector(object):
         .. _specificity: http://www.w3.org/TR/selectors/#specificity
 
         """
-        a, b, c = self._tree.specificity()
+        a, b, c = self.parsed_tree.specificity()
         if self.pseudo_element:
             c += 1
         return a, b, c
@@ -140,7 +140,7 @@ class Negation(object):
 
     def specificity(self):
         a1, b1, c1 = self.selector.specificity()
-        a2, b2, c2 = self.sub_selector.specificity()
+        a2, b2, c2 = self.subselector.specificity()
         return a1 + a2, b1 + b2, c1 + c2
 
 
@@ -276,7 +276,6 @@ def parse(css):
     try:
         return list(parse_selector_group(stream))
     except SelectorSyntaxError:
-        import sys
         e = sys.exc_info()[1]
         message = "%s at %s -> %r" % (
             e, stream.used, stream.peek())
@@ -357,10 +356,11 @@ def parse_simple_selector(stream, inside_negation=False):
             stream.next()
             result = parse_attrib(result, stream)
             next = stream.next()
-            if next != ']':
+            if next == ']':
+                continue
+            else:
                 raise SelectorSyntaxError(
-                    "] expected, got '%s'" % next)
-            continue
+                    "Expected ']', got '%s'" % next)
         elif peek == '::':
             stream.next()
             pseudo_element = stream.next_symbol()
@@ -373,36 +373,35 @@ def parse_simple_selector(stream, inside_negation=False):
                 # Any new pseudo-element must have two.
                 pseudo_element = ident
                 continue
-            if stream.peek() == '(':
-                stream.next()
-                stream.skip_whitespace()
-                if ident == 'not':
-                    if inside_negation:
-                        raise SelectorSyntaxError('Got nested :not()')
-                    argument, argument_pseudo_element = parse_simple_selector(
-                        stream, inside_negation=True)
-                    if argument_pseudo_element:
-                        raise SelectorSyntaxError(
-                            'Pseudo-elements are not allowed inside :not()')
-                else:
-                    peek = stream.peek()
-                    if isinstance(peek, (Symbol, String)):
-                        argument = stream.next()
-                    else:
-                        raise SelectorSyntaxError(
-                            "Expected argument, got '%s'" % peek)
-                stream.skip_whitespace()
-                next = stream.next()
-                if not next == ')':
-                    raise SelectorSyntaxError(
-                        "Expected ')', got '%s'" % next)
-                if ident == 'not':
-                    result = Negation(result, argument)
-                else:
-                    result = Function(result, ident, argument)
-            else:
+            if stream.peek() != '(':
                 result = Pseudo(result, ident)
-            continue
+                continue
+            stream.next()
+            stream.skip_whitespace()
+            if ident.lower() == 'not':
+                if inside_negation:
+                    raise SelectorSyntaxError('Got nested :not()')
+                argument, argument_pseudo_element = parse_simple_selector(
+                    stream, inside_negation=True)
+                if argument_pseudo_element:
+                    raise SelectorSyntaxError(
+                        'Pseudo-elements are not allowed inside :not()')
+                result = Negation(result, argument)
+            else:
+                peek = stream.peek()
+                if isinstance(peek, (Symbol, String)):
+                    argument = stream.next()
+                else:
+                    raise SelectorSyntaxError(
+                        "Expected argument, got '%s'" % peek)
+                result = Function(result, ident, argument)
+            stream.skip_whitespace()
+            next = stream.next()
+            if next == ')':
+                continue
+            else:
+                raise SelectorSyntaxError(
+                    "Expected ')', got '%s'" % next)
         else:
             raise SelectorSyntaxError(
                 "Expected selector, got '%s'" % peek)
@@ -442,16 +441,12 @@ def parse_attrib(selector, stream):
 
 def parse_series(s):
     """
-    Parses things like '1n+2', or 'an+b' generally, returning (a, b)
+    Parses things like '1n+2', or 'an+b' generally
+
+    :raises: :class:`ValueError`
+    :returns: :``(a, b)``
+
     """
-    if isinstance(s, Element):
-        s = s._format_element()
-    if not s or s == '*':
-        # Happens when there's nothing, which the CSS parser thinks of as *
-        return (0, 0)
-    if isinstance(s, int):
-        # Happens when you just get a number
-        return (0, s)
     if s == 'odd':
         return (2, 1)
     elif s == 'even':
@@ -459,7 +454,7 @@ def parse_series(s):
     elif s == 'n':
         return (1, 0)
     if 'n' not in s:
-        # Just a b
+        # Just b
         return (0, int(s))
     a, b = s.split('n', 1)
     if not a:
@@ -470,8 +465,6 @@ def parse_series(s):
         a = int(a)
     if not b:
         b = 0
-    elif b == '-' or b == '+':
-        b = int(b+'1')
     else:
         b = int(b)
     return (a, b)
@@ -596,22 +589,18 @@ _illegal_symbol = re.compile(r'[^\w\\-]', re.UNICODE)
 def tokenize_symbol(s, pos):
     start = pos
     match = _illegal_symbol.search(s, pos=pos)
-    if not match:
-        # Goes to end of s
-        return s[start:], len(s)
-    if match.start() == pos:
-        raise SelectorSyntaxError(
-            "Unexpected symbol: %r" % s[pos])
-    if not match:
-        result = s[start:]
-        pos = len(s)
-    else:
+    if match:
+        if match.start() == pos:
+            raise SelectorSyntaxError(
+                "Unexpected symbol: %r" % s[pos])
         result = s[start:match.start()]
         pos = match.start()
+    else:
+        result = s[start:]
+        pos = len(s)
     try:
         result = result.encode('ASCII', 'backslashreplace').decode('unicode_escape')
     except UnicodeDecodeError:
-        import sys
         e = sys.exc_info()[1]
         raise SelectorSyntaxError(
             "Bad symbol %r: %s" % (result, e))
@@ -644,9 +633,6 @@ class TokenStream(object):
             except StopIteration:
                 return None
 
-    def __iter__(self):
-        return iter(self.next, None)
-
     def peek(self):
         if not self._peeking:
             try:
