@@ -20,7 +20,7 @@ import sys
 import operator
 import unittest
 
-from lxml import html
+from lxml import etree, html
 from cssselect import (parse, GenericTranslator, HTMLTranslator,
                        SelectorSyntaxError, ExpressionError)
 from cssselect.parser import tokenize, parse_series
@@ -47,7 +47,7 @@ class TestCssselect(unittest.TestCase):
             selectors = parse(css)
             for selector in selectors:
                 assert selector.pseudo_element is None
-            return [repr(selector._tree).replace("(u'", "('")
+            return [repr(selector.parsed_tree).replace("(u'", "('")
                     for selector in selectors]
 
         def parse_many(first, *others):
@@ -132,7 +132,7 @@ class TestCssselect(unittest.TestCase):
             result = []
             for selector in parse(css):
                 result.append((
-                    repr(selector._tree).replace("(u'", "('"),
+                    repr(selector.parsed_tree).replace("(u'", "('"),
                     selector.pseudo_element))
             return result
 
@@ -187,6 +187,13 @@ class TestCssselect(unittest.TestCase):
         assert specificity('[baz^="4"]') == (0, 1, 0)
         assert specificity('#lipsum') == (1, 0, 0)
 
+        assert specificity(':not(*)') == (0, 0, 0)
+        assert specificity(':not(foo)') == (0, 0, 1)
+        assert specificity(':not(.foo)') == (0, 1, 0)
+        assert specificity(':not([foo])') == (0, 1, 0)
+        assert specificity(':not(:empty)') == (0, 1, 0)
+        assert specificity(':not(#foo)') == (1, 0, 0)
+
         assert specificity('foo:empty') == (0, 1, 1)
         assert specificity('foo:before') == (0, 0, 2)
         assert specificity('foo::before') == (0, 0, 2)
@@ -263,6 +270,22 @@ class TestCssselect(unittest.TestCase):
             "Expected selector, got 'a' at "
             "[Token('[', 0), Symbol('href', 1), Token(']', 5)]"
             " -> Symbol('a', 6)")
+        assert get_error('[rel=stylesheet]') == None
+        assert get_error('[rel:stylesheet]') == (
+            "Operator expected, got ':' at [Token('[', 0), Symbol('rel', 1), "
+            "Token(':', 4)] -> Symbol('stylesheet', 5)")
+        assert get_error('[rel=stylesheet') == (
+            "Expected ']', got 'None' at [Token('[', 0), Symbol('rel', 1), "
+            "Token('=', 4), Symbol('stylesheet', 5)] -> None")
+        assert get_error(':lang(fr)') == None
+        assert get_error(':lang(fr') == (
+            "Expected ')', got 'None' at [Token(':', 0), Symbol('lang', 1), "
+            "Token('(', 5), Symbol('fr', 6)] -> None")
+        assert get_error(':contains("foo') == (
+            "Expected closing \" for string in: 'foo' at "
+            "[Token(':', 0), Symbol('contains', 1), Token('(', 9)] -> None")
+        assert get_error('foo!') == (
+            "Unexpected symbol: '!' at [Symbol('foo', 0)] -> None")
 
         # Mis-placed pseudo-elements
         assert get_error('a:before:empty') == (
@@ -277,89 +300,112 @@ class TestCssselect(unittest.TestCase):
             "Pseudo-elements are not allowed inside :not() at "
             "[Token(':', 0), Symbol('not', 1), Token('(', 4), Token(':', 5),"
             " Symbol('before', 6)] -> Token(')', 12)")
-
+        assert get_error(':not(:not(a))') == (
+            "Got nested :not() at [Token(':', 0), Symbol('not', 1), "
+            "Token('(', 4), Token(':', 5), Symbol('not', 6), Token('(', 9)]"
+            " -> Symbol('a', 10)")
 
     def test_translation(self):
         def xpath(css):
             return str(GenericTranslator().css_to_xpath(css, prefix=''))
 
         assert xpath('*') == "*"
-        assert xpath('E') == "e"
-        assert xpath('E[foo]') == "e[@foo]"
-        assert xpath('E[foo="bar"]') == "e[@foo = 'bar']"
-        assert xpath('E[foo~="bar"]') == (
+        assert xpath('e') == "e"
+        assert xpath('*|e') == "e"
+        assert xpath('e|f') == "e:f"
+        assert xpath('e[foo]') == "e[@foo]"
+        assert xpath('e[foo|bar]') == "e[@foo:bar]"
+        assert xpath('e[foo="bar"]') == "e[@foo = 'bar']"
+        assert xpath('e[foo~="bar"]') == (
             "e[@foo and contains("
                "concat(' ', normalize-space(@foo), ' '), ' bar ')]")
-        assert xpath('E[foo^="bar"]') == (
+        assert xpath('e[foo^="bar"]') == (
             "e[@foo and starts-with(@foo, 'bar')]")
-        assert xpath('E[foo$="bar"]') == (
+        assert xpath('e[foo$="bar"]') == (
             "e[@foo and substring(@foo, string-length(@foo)-2) = 'bar']")
-        assert xpath('E[foo*="bar"]') == (
+        assert xpath('e[foo*="bar"]') == (
             "e[@foo and contains(@foo, 'bar')]")
-        assert xpath('E[hreflang|="en"]') == (
+        assert xpath('e[hreflang|="en"]') == (
             "e[@hreflang and ("
                "@hreflang = 'en' or starts-with(@hreflang, 'en-'))]")
-        assert xpath('E:nth-child(1)') == (
+        assert xpath('e:nth-child(1)') == (
             "*/*[name() = 'e' and (position() = 1)]")
-        assert xpath('E:nth-last-child(1)') == (
+        assert xpath('e:nth-last-child(1)') == (
             "*/*[name() = 'e' and (position() = last() - 1)]")
-        assert xpath('E:nth-last-child(2n+2)') == (
+        assert xpath('e:nth-last-child(2n+2)') == (
             "*/*[name() = 'e' and ("
                "(position() +2) mod -2 = 0 and position() < (last() -2))]")
-        assert xpath('E:nth-of-type(1)') == (
+        assert xpath('e:nth-of-type(1)') == (
             "*/e[position() = 1]")
-        assert xpath('E:nth-last-of-type(1)') == (
+        assert xpath('e:nth-last-of-type(1)') == (
             "*/e[position() = last() - 1]")
-        assert xpath('E:nth-last-of-type(1)') == (
+        assert xpath('e:nth-last-of-type(1)') == (
             "*/e[position() = last() - 1]")
-        assert xpath('div E:nth-last-of-type(1) .aclass') == (
+        assert xpath('div e:nth-last-of-type(1) .aclass') == (
             "div/descendant-or-self::*/e[position() = last() - 1]"
                "/descendant-or-self::*/*[@class and contains("
                "concat(' ', normalize-space(@class), ' '), ' aclass ')]")
-        assert xpath('E:first-child') == (
+        assert xpath('e:first-child') == (
             "*/*[name() = 'e' and (position() = 1)]")
-        assert xpath('E:last-child') == (
+        assert xpath('e:last-child') == (
             "*/*[name() = 'e' and (position() = last())]")
-        assert xpath('E:first-of-type') == (
+        assert xpath('e:first-of-type') == (
             "*/e[position() = 1]")
-        assert xpath('E:last-of-type') == (
+        assert xpath('e:last-of-type') == (
             "*/e[position() = last()]")
-        assert xpath('E:only-child') == (
+        assert xpath('e:only-child') == (
             "*/*[name() = 'e' and (last() = 1)]")
-        assert xpath('E:only-of-type') == (
+        assert xpath('e:only-of-type') == (
             "e[last() = 1]")
-        assert xpath('E:empty') == (
+        assert xpath('e:empty') == (
             "e[not(*) and not(normalize-space())]")
-        assert xpath('E:root') == (
+        assert xpath('e:root') == (
             "e[not(parent::*)]")
-        assert xpath('E:contains("foo")') == (
+        assert xpath('e:hover') == (
+            "e[0]")  # never matches
+        assert xpath('e:contains("foo")') == (
             "e[contains(string(.), 'foo')]")
-        assert xpath('E:contains(foo)') == (
+        assert xpath('e:contains(foo)') == (
             "e[contains(string(.), 'foo')]")
-        assert xpath('E.warning') == (
+        assert xpath('e.warning') == (
             "e[@class and contains("
                "concat(' ', normalize-space(@class), ' '), ' warning ')]")
-        assert xpath('E#myid') == (
+        assert xpath('e#myid') == (
             "e[@id = 'myid']")
-        assert xpath('E:not(:nth-child(odd))') == (
+        assert xpath('e:not(:nth-child(odd))') == (
             "e[not((position() -1) mod 2 = 0 and position() >= 1)]")
-        assert xpath('E F') == (
+        assert xpath('e:not(*)') == (
+            "e[0]")  # never matches
+        assert xpath('e f') == (
             "e/descendant-or-self::*/f")
-        assert xpath('E > F') == (
+        assert xpath('e > f') == (
             "e/f")
-        assert xpath('E + F') == (
+        assert xpath('e + f') == (
             "e/following-sibling::*[name() = 'f' and (position() = 1)]")
-        assert xpath('E ~ F') == (
+        assert xpath('e ~ f') == (
             "e/following-sibling::f")
         assert xpath('div#container p') == (
             "div[@id = 'container']/descendant-or-self::*/p")
-        self.assertRaises(ExpressionError, xpath, 'p *:only-of-type')
+        self.assertRaises(ExpressionError, xpath, ':first-of-type')
+        self.assertRaises(ExpressionError, xpath, ':only-of-type')
+        self.assertRaises(ExpressionError, xpath, ':last-of-type')
+        self.assertRaises(ExpressionError, xpath, ':nth-of-type(1)')
+        self.assertRaises(ExpressionError, xpath, ':nth-last-of-type(1)')
+        self.assertRaises(ExpressionError, xpath, ':lang(fr)')
+        self.assertRaises(ExpressionError, xpath, ':nth-child(n-)')
+        self.assertRaises(ExpressionError, xpath, ':after')
+        self.assertRaises(ExpressionError, xpath, ':lorem-ipsum')
+        self.assertRaises(ExpressionError, xpath, ':lorem(ipsum)')
+        self.assertRaises(ExpressionError, xpath, '::lorem-ipsum')
+        self.assertRaises(TypeError, GenericTranslator().css_to_xpath, 4)
+        self.assertRaises(TypeError, GenericTranslator().selector_to_xpath,
+            'foo')
 
     def test_unicode(self):
-        if sys.version_info[0] >= 3:
-            css = '.a\xc1b'
-        else:
+        if sys.version_info[0] < 3:
             css = '.a\xc1b'.decode('ISO-8859-1')
+        else:
+            css = '.a\xc1b'
 
         xpath = GenericTranslator().css_to_xpath(css)
         assert css[1:] in xpath
@@ -398,10 +444,14 @@ class TestCssselect(unittest.TestCase):
         assert parse_series('even') == (2, 0)
         assert parse_series('3n') == (3, 0)
         assert parse_series('n') == (1, 0)
+        assert parse_series('+n') == (1, 0)
+        assert parse_series('-n') == (-1, 0)
         assert parse_series('5') == (0, 5)
+        self.assertRaises(ValueError, parse_series, 'foo')
+        self.assertRaises(ValueError, parse_series, 'n+')
 
     def test_select(self):
-        document = html.document_fromstring(HTML_IDS)
+        document = etree.fromstring(HTML_IDS)
         sort_key = dict(
             (el, count) for count, el in enumerate(document.getiterator())
         ).__getitem__
@@ -426,12 +476,17 @@ class TestCssselect(unittest.TestCase):
             return result
 
         all_ids = pcss('*')
-        assert all_ids[:4] == ['html', 'nil', 'nil', 'outer-div']
+        assert all_ids[:6] == [
+            'html', 'nil', 'link-href', 'link-nohref', 'nil', 'outer-div']
         assert all_ids[-1:] == ['foobar-span']
         assert pcss('div') == ['outer-div', 'li-div', 'foobar-div']
+        assert pcss('DIV', html_only=True) == [
+            'outer-div', 'li-div', 'foobar-div']  # case-insensitive in HTML
         assert pcss('div div') == ['li-div']
         assert pcss('div, div div') == ['outer-div', 'li-div', 'foobar-div']
         assert pcss('a[name]') == ['name-anchor']
+        assert pcss('a[NAme]', html_only=True) == [
+            'name-anchor'] # case-insensitive in HTML:
         assert pcss('a[rel]') == ['tag-anchor', 'nofollow-anchor']
         assert pcss('a[rel="tag"]') == ['tag-anchor']
         assert pcss('a[href*="localhost"]') == ['tag-anchor']
@@ -441,7 +496,7 @@ class TestCssselect(unittest.TestCase):
         assert pcss('div[foobar~="bc"]', 'div[foobar~="cde"]') == [
             'foobar-div']
         assert pcss('div[foobar~="cd"]') == []
-        assert pcss('*[lang|="en"]', '*[lang|="en-US"]') == ['second-li']
+        assert pcss('*[lang|="en"]', '[lang|="en-US"]') == ['second-li']
         assert pcss('*[lang|="e"]') == []
         assert pcss('li:nth-child(3)') == ['third-li']
         assert pcss('li:nth-child(10)') == []
@@ -466,17 +521,16 @@ class TestCssselect(unittest.TestCase):
         assert pcss('ol:nth-last-of-type(1)') == ['first-ol']
         assert pcss('span:only-child') == ['foobar-span']
         assert pcss('li div:only-child') == ['li-div']
-        assert pcss('div *:only-child') == [
-            'li-div', 'checkbox-disabled', 'foobar-span']
+        assert pcss('div *:only-child') == ['li-div', 'foobar-span']
         self.assertRaises(ExpressionError, pcss, 'p *:only-of-type')
         self.assertRaises(ExpressionError, pcss, 'p:lang(fr)')
         assert pcss('p:only-of-type') == ['paragraph']
-        assert pcss('a:empty') == ['name-anchor']
+        assert pcss('a:empty', 'a:EMpty') == ['name-anchor']
         assert pcss('li:empty') == [
             'third-li', 'fourth-li', 'fifth-li', 'sixth-li', 'seventh-li']
         assert pcss(':root', 'html:root') == ['html']
         assert pcss('li:root', '* :root') == []
-        assert pcss('*:contains("link")') == [
+        assert pcss('*:contains("link")', ':CONtains("link")') == [
             'html', 'nil', 'outer-div', 'tag-anchor', 'nofollow-anchor']
         assert pcss('*:contains("LInk")') == []  # case sensitive
         assert pcss('*:contains("e")') == [
@@ -488,7 +542,6 @@ class TestCssselect(unittest.TestCase):
         assert pcss('ol *.c', 'ol li.c', 'li ~ li.c', 'ol > li.c') == [
             'third-li', 'fourth-li']
         assert pcss('#first-li', 'li#first-li', '*#first-li') == ['first-li']
-        # Need some tests of :not()']
         assert pcss('li div', 'li > div', 'div div') == ['li-div']
         assert pcss('div > div') == []
         assert pcss('div>.c', 'div > .c') == ['first-ol']
@@ -499,14 +552,28 @@ class TestCssselect(unittest.TestCase):
         assert pcss('ol#first-ol *:last-child') == ['li-div', 'seventh-li']
         assert pcss('#outer-div:first-child') == ['outer-div']
         assert pcss('#outer-div :first-child') == [
-            'name-anchor', 'first-li', 'li-div', 'p-b', 'checkbox-disabled']
+            'name-anchor', 'first-li', 'li-div', 'p-b',
+            'checkbox-fieldset-disabled', 'area-href']
         assert pcss('a[href]') == ['tag-anchor', 'nofollow-anchor']
-        assert pcss(':link', html_only=True) == pcss('a[href]')
-        assert pcss(':checked', html_only=True) == ['checkbox-checked']
-        assert pcss(':disabled', html_only=True) == [
-            'fieldset', 'checkbox-disabled']
+        assert pcss(':not(*)') == []
+        assert pcss('a:not([href])') == ['name-anchor']
+        assert pcss('ol :Not(li[class])') == [
+            'first-li', 'second-li', 'li-div',
+            'fifth-li', 'sixth-li', 'seventh-li']
+
+        # HTML-specific
+        assert pcss(':link', html_only=True) == [
+            'link-href', 'tag-anchor', 'nofollow-anchor', 'area-href']
+        assert pcss(':visited', html_only=True) == []
         assert pcss(':enabled', html_only=True) == [
-            'checkbox-unchecked', 'checkbox-checked']
+            'link-href', 'tag-anchor', 'nofollow-anchor',
+            'checkbox-unchecked', 'text-checked', 'checkbox-checked',
+            'area-href']
+        assert pcss(':disabled', html_only=True) == [
+            'checkbox-disabled', 'checkbox-disabled-checked', 'fieldset',
+            'checkbox-fieldset-disabled']
+        assert pcss(':checked', html_only=True) == [
+            'checkbox-checked', 'checkbox-disabled-checked']
 
     def test_select_shakespeare(self):
         document = html.document_fromstring(HTML_SHAKESPEARE)
@@ -582,7 +649,10 @@ class TestCssselect(unittest.TestCase):
         assert count('div[class~=dialog]') == 51 # ? Seems right
 
 HTML_IDS = '''
-<html id="html"><head></head><body>
+<html id="html"><head>
+  <link id="link-href" href="foo" />
+  <link id="link-nohref" />
+</head><body>
 <div id="outer-div">
  <a id="name-anchor" name="foo"></a>
  <a id="tag-anchor" rel="tag" href="http://localhost/foo">link</a>
@@ -604,14 +674,25 @@ c"></li>
  <p id="paragraph">
    <b id="p-b">hi</b> <em id="p-em">there</em>
    <b id="p-b2">guy</b>
-   <input type="checkbox" id="checkbox-unchecked">
-   <input type="checkbox" id="checkbox-checked" checked="checked">
+   <input type="checkbox" id="checkbox-unchecked" />
+   <input type="checkbox" id="checkbox-disabled" disabled="" />
+   <input type="text" id="text-checked" checked="checked" />
+   <input type="hidden" />
+   <input type="hidden" disabled="disabled" />
+   <input type="checkbox" id="checkbox-checked" checked="checked" />
+   <input type="checkbox" id="checkbox-disabled-checked"
+          disabled="disabled" checked="checked" />
    <fieldset id="fieldset" disabled="disabled">
-     <input type="checkbox" id="checkbox-disabled">
+     <input type="checkbox" id="checkbox-fieldset-disabled" />
+     <input type="hidden" />
    </fieldset>
  </p>
  <ol id="second-ol">
  </ol>
+ <map name="dummymap">
+   <area shape="circle" coords="200,250,25" href="foo.html" id="area-href" />
+   <area shape="default" id="area-nohref" />
+ </map>
 </div>
 <div id="foobar-div" foobar="ab bc
 cde"><span id="foobar-span"></span></div>
