@@ -11,15 +11,16 @@
 
 """
 
+import sys
 import re
+
 from cssselect.parser import parse, parse_series, SelectorError
 
 
-try:
+if sys.version_info[0] < 3:
     _basestring = basestring
     _unicode = unicode
-except NameError:
-    # Python 3
+else:
     _basestring = str
     _unicode = str
 
@@ -36,7 +37,6 @@ class XPathExpr(object):
         self.path = path
         self.element = element
         self.condition = condition
-        self.star_prefix = star_prefix
 
     def __str__(self):
         path =  _unicode(self.path) + _unicode(self.element)
@@ -64,20 +64,15 @@ class XPathExpr(object):
 
     def add_star_prefix(self):
         """
-        Adds a /* prefix if there is no prefix.  This is when you need
-        to keep context's constrained to a single parent.
+        Append '*/' to the path to keep the context constrained
+        to a single parent.
         """
-        if self.path:
-            self.path += '*/'
-        else:
-            self.path = '*/'
-        self.star_prefix = True
+        self.path += '*/'
 
     def join(self, combiner, other):
         path = _unicode(self) + combiner
-        # We don't need a star prefix if we are joining to this other
-        # prefix; so we'll get rid of it
-        if not(other.star_prefix and other.path == '*/'):
+        # Any "star prefix" is redundant when joining.
+        if other.path != '*/':
             path += other.path
         self.path = path
         self.element = other.element
@@ -93,6 +88,10 @@ split_at_single_quotes = re.compile("('+)").split
 class GenericTranslator(object):
     """
     Translator for "generic" XML documents.
+
+    Everything is case-sensitive, no assumption is made on the meaning
+    of element names and attribute names.
+
     """
     combinator_mapping = {
         ' ': 'descendant',
@@ -116,6 +115,24 @@ class GenericTranslator(object):
     #: http://www.w3.org/TR/selectors/#id-selectors
     id_attribute = 'id'
 
+    #: The case sensitivity of document language element names,
+    #: attribute names, and attribute values in selectors depends
+    #: on the document language.
+    #: http://www.w3.org/TR/selectors/#casesens
+    #:
+    #: When a document language defines one of these as case-insensitive,
+    #: cssselect assumes that the document parser makes the parsed values
+    #: lower-case. Making the selector lower-case too makes the comparaison
+    #: case-insensitive.
+    #:
+    #: In HTML, element names and attributes names (but not attribute values)
+    #: are case-insensitive. All of lxml.html, html5lib, BeautifulSoup4
+    #: and HTMLParser make them lower-case in their parse result, so
+    #: the assumption holds.
+    lower_case_element_names = False
+    lower_case_attribute_names = False
+    lower_case_attribute_values = False
+
     def css_to_xpath(self, css, prefix='descendant-or-self::'):
         """Translate a *group of selectors* to XPath.
 
@@ -156,7 +173,10 @@ class GenericTranslator(object):
             The equivalent XPath 1.0 expression as an Unicode string.
 
         """
-        return (prefix or '') + _unicode(self.xpath(selector._tree))
+        tree = getattr(selector, 'parsed_tree', None)
+        if not tree:
+            raise TypeError('Expected a parsed selector, got %r' % (selector,))
+        return (prefix or '') + _unicode(self.xpath(tree))
 
     @staticmethod
     def xpath_literal(s):
@@ -175,9 +195,7 @@ class GenericTranslator(object):
     def xpath(self, parsed_selector):
         """Translate any parsed selector object."""
         type_name = type(parsed_selector).__name__
-        method = getattr(self, 'xpath_%s' % type_name.lower(), None)
-        if not method:
-            raise TypeError('Expected a parsed selector, got %s' % type_name)
+        method = getattr(self, 'xpath_%s' % type_name.lower())
         return method(parsed_selector)
 
 
@@ -185,10 +203,7 @@ class GenericTranslator(object):
 
     def xpath_combinedselector(self, combined):
         """Translate a combined selector."""
-        combinator = self.combinator_mapping.get(combined.combinator)
-        if not combinator:
-            raise ExpressionError(
-                "Unknown combinator: %r" % combined.combinator)
+        combinator = self.combinator_mapping[combined.combinator]
         method = getattr(self, 'xpath_%s_combinator' % combinator)
         return method(self.xpath(combined.selector),
                       self.xpath(combined.subselector))
@@ -197,11 +212,14 @@ class GenericTranslator(object):
         xpath = self.xpath(negation.selector)
         sub_xpath = self.xpath(negation.subselector)
         sub_xpath.add_name_test()
-        return xpath.add_condition('not(%s)' % sub_xpath.condition)
+        if sub_xpath.condition:
+            return xpath.add_condition('not(%s)' % sub_xpath.condition)
+        else:
+            return xpath.add_condition('0')
 
     def xpath_function(self, function):
         """Translate a functional pseudo-class."""
-        method = 'xpath_%s_function' % function.name.replace('-', '_')
+        method = 'xpath_%s_function' % function.name.replace('-', '_').lower()
         method = getattr(self, method, None)
         if not method:
             raise ExpressionError(
@@ -210,7 +228,7 @@ class GenericTranslator(object):
 
     def xpath_pseudo(self, pseudo):
         """Translate a pseudo-class."""
-        method = 'xpath_%s_pseudo' % pseudo.ident.replace('-', '_')
+        method = 'xpath_%s_pseudo' % pseudo.ident.replace('-', '_').lower()
         method = getattr(self, method, None)
         if not method:
             # TODO: better error message for pseudo-elements?
@@ -221,17 +239,21 @@ class GenericTranslator(object):
 
     def xpath_attrib(self, selector):
         """Translate an attribute selector."""
-        operator = self.attribute_operator_mapping.get(selector.operator)
-        if not operator:
-            raise ExpressionError(
-                "Unknown attribute operator: %r" % selector.operator)
+        operator = self.attribute_operator_mapping[selector.operator]
         method = getattr(self, 'xpath_attrib_%s' % operator)
-        # FIXME: what if attrib is *?
-        if selector.namespace == '*':
-            name = '@' + selector.attrib
+        if self.lower_case_attribute_names:
+            name = selector.attrib.lower()
         else:
-            name = '@%s:%s' % (selector.namespace, selector.attrib)
-        return method(self.xpath(selector.selector), name, selector.value)
+            name = selector.attrib
+        if selector.namespace == '*':
+            name = '@' + name
+        else:
+            name = '@%s:%s' % (selector.namespace, name)
+        if self.lower_case_attribute_values:
+            value = selector.value.lower()
+        else:
+            value = selector.value
+        return method(self.xpath(selector.selector), name, value)
 
     def xpath_class(self, class_selector):
         """Translate a class selector."""
@@ -243,16 +265,18 @@ class GenericTranslator(object):
     def xpath_hash(self, id_selector):
         """Translate an ID selector."""
         xpath = self.xpath(id_selector.selector)
-        return xpath.add_condition('@%s = %s' % (
-            self.id_attribute, self.xpath_literal(id_selector.id)))
+        return self.xpath_attrib_equals(xpath, '@id', id_selector.id)
 
     def xpath_element(self, selector):
         """Translate a type or universal selector."""
-        if selector.namespace == '*':
+        if self.lower_case_element_names:
             element = selector.element.lower()
         else:
-            # FIXME: Should we lowercase here?
-            element = '%s:%s' % (selector.namespace, selector.element)
+            element = selector.element
+        if selector.namespace != '*':
+            # Namespace prefixes are case-sensitive.
+            # http://www.w3.org/TR/css3-namespace/#prefixes
+            element = '%s:%s' % (selector.namespace, element)
         return XPathExpr(element=element)
 
 
@@ -281,10 +305,10 @@ class GenericTranslator(object):
 
     def xpath_nth_child_function(self, xpath, function, last=False,
                                  add_name_test=True):
-        a, b = parse_series(function.arguments)
-        if not a and not b and not last:
-            # a=0 means nothing is returned...
-            return xpath.add_condition('false() and position() = 0')
+        try:
+            a, b = parse_series(function.arguments)
+        except ValueError:
+            raise ExpressionError("Invalid series: '%r'" % function.arguments)
         if add_name_test:
             xpath.add_name_test()
         xpath.add_star_prefix()
@@ -458,16 +482,38 @@ class GenericTranslator(object):
 
 class HTMLTranslator(GenericTranslator):
     """
-    Translator for HTML documents.
+    Translator for (X)HTML documents.
+
+    Has a more useful implementation of some pseudo-classes based on
+    HTML-specific element names and attribute names, as described in
+    the `HTML5 specification`_. It assumes no-quirks mode.
+    The API is the same as :class:`GenericTranslator`.
+
+    .. _HTML5 specification: http://www.w3.org/TR/html5/links.html#selectors
+
+    :param xhtml:
+        If false (the default), element names and attribute names
+        are case-insensitive.
+
     """
+    def __init__(self, xhtml=False):
+        self.xhtml = xhtml  # Might be useful for sub-classes?
+        if not xhtml:
+            # See their definition in GenericTranslator.
+            self.lower_case_element_names = True
+            self.lower_case_attribute_names = True
+
     def xpath_checked_pseudo(self, xpath):
         # FIXME: is this really all the elements?
         return xpath.add_condition(
             "(@selected and name(.) = 'option') or "
-            "(@checked and name(.) = 'input')")
+            "(@checked "
+                "and (name(.) = 'input' or name(.) = 'command')"
+                "and (@type = 'checkbox' or @type = 'radio'))")
 
     def xpath_link_pseudo(self, xpath):
-        return xpath.add_condition("@href and name(.) = 'a'")
+        return xpath.add_condition("@href and "
+            "(name(.) = 'a' or name(.) = 'link' or name(.) = 'area')")
 
     # Links are never visited, the implementation for :visited is the same
     # as in GenericTranslator
@@ -478,11 +524,10 @@ class HTMLTranslator(GenericTranslator):
         (
             @disabled and
             (
-                name(.) = 'input' or
+                (name(.) = 'input' and @type != 'hidden') or
                 name(.) = 'button' or
                 name(.) = 'select' or
                 name(.) = 'textarea' or
-                name(.) = 'keygen' or
                 name(.) = 'command' or
                 name(.) = 'fieldset' or
                 name(.) = 'optgroup' or
@@ -490,11 +535,10 @@ class HTMLTranslator(GenericTranslator):
             )
         ) or (
             (
-                name(.) = 'input' or
+                (name(.) = 'input' and @type != 'hidden') or
                 name(.) = 'button' or
                 name(.) = 'select' or
-                name(.) = 'textarea' or
-                name(.) = 'keygen'
+                name(.) = 'textarea'
             )
             and ancestor::fieldset[@disabled]
         )
@@ -506,23 +550,36 @@ class HTMLTranslator(GenericTranslator):
         # http://www.w3.org/TR/html5/section-index.html#attributes-1
         return xpath.add_condition('''
         (
+            @href and (
+                name(.) = 'a' or
+                name(.) = 'link' or
+                name(.) = 'area'
+            )
+        ) or (
             (
                 name(.) = 'command' or
                 name(.) = 'fieldset' or
-                name(.) = 'optgroup' or
-                name(.) = 'option'
+                name(.) = 'optgroup'
             )
             and not(@disabled)
         ) or (
             (
-                name(.) = 'input' or
+                (name(.) = 'input' and @type != 'hidden') or
                 name(.) = 'button' or
                 name(.) = 'select' or
                 name(.) = 'textarea' or
                 name(.) = 'keygen'
             )
             and not (@disabled or ancestor::fieldset[@disabled])
+        ) or (
+            name(.) = 'option' and not(
+                @disabled or ancestor::optgroup[@disabled]
+            )
         )
         ''')
-        # FIXME: in the second half, add "and is not a descendant of that
-        # fieldset element's first legend element child, if any."
+        # FIXME: ... or "li elements that are children of menu elements,
+        # and that have a child element that defines a command, if the first
+        # such element's Disabled State facet is false (not disabled)".
+        # FIXME: after ancestor::fieldset[@disabled], add "and is not a
+        # descendant of that fieldset element's first legend element child,
+        # if any."
